@@ -14,7 +14,8 @@ import           Control.Applicative
 import           Control.Arrow        (second)
 import           Control.Monad.Catch
 import           Data.Aeson
-import           Data.Aeson.Types (explicitParseFieldMaybe, Parser)
+import           Data.Aeson.Types     (Parser, explicitParseFieldMaybe)
+import           Data.Binary          (Binary)
 import qualified Data.ByteString      as BS
 import qualified Data.ByteString.Lazy as BSL
 import           Data.Maybe
@@ -36,15 +37,15 @@ import           URI.ByteString.Aeson ()
 -- | Query Parameter Representation
 data OAuth2 = OAuth2 {
       oauthClientId            :: Text
-    , oauthClientSecret        :: Text
+    , oauthClientSecret        :: Maybe Text
     , oauthOAuthorizeEndpoint  :: URI
     , oauthAccessTokenEndpoint :: URI
     , oauthCallback            :: Maybe URI
     } deriving (Show, Eq)
 
-newtype AccessToken = AccessToken { atoken :: Text } deriving (Show, FromJSON, ToJSON)
-newtype RefreshToken = RefreshToken { rtoken :: Text } deriving (Show, FromJSON, ToJSON)
-newtype IdToken = IdToken { idtoken :: Text } deriving (Show, FromJSON, ToJSON)
+newtype AccessToken = AccessToken { atoken :: Text } deriving (Binary, Eq, Show, FromJSON, ToJSON)
+newtype RefreshToken = RefreshToken { rtoken :: Text } deriving (Binary, Eq, Show, FromJSON, ToJSON)
+newtype IdToken = IdToken { idtoken :: Text } deriving (Binary, Eq, Show, FromJSON, ToJSON)
 newtype ExchangeToken = ExchangeToken { extoken :: Text } deriving (Show, FromJSON, ToJSON)
 
 
@@ -58,11 +59,13 @@ data OAuth2Token = OAuth2Token {
     , expiresIn    :: Maybe Int
     , tokenType    :: Maybe Text
     , idToken      :: Maybe IdToken
-    } deriving (Show, Generic)
+    } deriving (Eq, Show, Generic)
+
+instance Binary OAuth2Token
 
 parseIntFlexible :: Value -> Parser Int
 parseIntFlexible (String s) = pure . read $ unpack s
-parseIntFlexible v = parseJSON v
+parseIntFlexible v          = parseJSON v
 
 -- | Parse JSON data into 'OAuth2Token'
 instance FromJSON OAuth2Token where
